// @ myth - server - beast.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk
//
// Definition of rutines used by the beast object.

// @@ include

#include "server/beast.h"
#include "common/error.h"
#include "common/parse.h"
#include "common/page.h"
#include "server/watch.h"
#include "server/level.h"
#include "server/rnd.h"
#include "server/corpse.h"
#include "server/player.h"
#include "server/ring.h"
#include "common/format.h"

// @@ BEAST

// @@@ cont BEAST::UseItem(LEVEL* level, symbol_type symbol)

cont BEAST::UseItem(LEVEL* level, symbol_type symbol)
{
    OBJECT_LIST list;
    SPECIFY_SYMBOL specify(symbol);
    pack.RemoveHalf(&specify, &list);
    
    if(list.length() > 0)
    {
	list.reset();
	
	if(!pack.worn(list.current()))
	{
	    if(list.current()->count() > 1L)
		list.current()->split(1L)->apply(level, this);
	    else
	    {
		list.current()->apply(level, this);
		list.remove();
	    }
	}
	pack.add(&list); // Other items to the pack
    }
    return MayMove() ? CONTINUE : DONE;
}

// @@@ bool BEAST::IsEnemy(OBJECT* object)

bool BEAST::IsEnemy(OBJECT* object)
{ return enemies.member(object) || object->IsPlayer(); }

// @@@ bool BEAST::IsPicking(OBJECT* object)

bool BEAST::IsPicking(OBJECT* object)
{
    if(object->worth(this) < MinWorth)
	return FALSE;
    else if(IsWanted(object))
	return TRUE;
    else if(knowledge(object) > 0)
	return TRUE;
    else
	return FALSE;
}

// @@@ bool BEAST::UseThing()

bool BEAST::UseThing()
{ return UseFood() || UseWeapon() || UseMagic(); }

// @@@ MONSTER* BEAST::enemy(LEVEL* level)

MONSTER* BEAST::enemy(LEVEL* level)
{
    MONSTER* monster = level->FindEnemy(this);

    if(monster == NULL)
	return this;
    else
	return monster;
}

// @@@ STRING BEAST::hurt(int power, OBJECT* object,

STRING BEAST::hurt(int power, OBJECT* object, MONSTER* monster, LEVEL* level)
{
    if(monster != NULL)
	enemies.add(monster);

    return MONSTER::hurt(power, object, monster, level);
}

// @@@ cont BEAST::play(LEVEL* level)
//
// A real MONSTER procedure. Any program need at least one of these :-)
//
// Contain the code for most monster actions, but see also SpecialAction()

cont BEAST::play(LEVEL* level)
{
    if(!MayMove() || dead())
	return DONE;

    // * Update counters &c
    
    SpendTime(level);
    UseMovePoint(5);	// Just in case...
    
    // * Pick Up

    if(UseThing() && LightLoad())
    {
	OBJECT_LIST list;
	SPECIFY_PICK specify(this);
  	BUNCH* bunch = level->create(position());
	bunch->remove(&specify, &list);
	if(list.length() > 0)
	{
	    
	    UseMovePoint(list.length() * 10);
	    pack.add(&list);

	    // Tell the world.
	    if(list.length() == 1)
	    {
		list.reset();
		level->WatchMonster(this, "%s picks %s up.", list.current());
	    }
	    else
		level->WatchMonster(this, "%s picks up several items.");

	    return MayMove() ? CONTINUE : DONE;
	}
    }
	
    // * Use or Throw Away
    
    if(MayFight())
    {
	switch(rnd(1, 25))
	{
	case 1:			// Select the best weapon and armor
	    if(UseWeapon())
		return UseItem(level, SYMBOL_WEAPON);
	    break;
	case 2:
	    if(UseWeapon())
		return UseItem(level, SYMBOL_ARMOR);
	    break;
	case 3:
	    if(UseMagic())
		return UseItem(level, SYMBOL_RING);
	    break;
	case 4:
	    if(UseMagic() && (body->wounded() || rnd.chance(1, 20)))
		return UseItem(level, SYMBOL_SCROLL);
	    break;
	case 5:
	    if(UseMagic() && (body->wounded() || rnd.chance(1, 10)))
		return UseItem(level, SYMBOL_POTION);
	    break;
	case 6:
	    if(UseMagic() && (body->wounded() || rnd.chance(1, 5)))
		return UseItem(level, SYMBOL_WAND);
	    break;
	case 7:
	    if(UseFood() && body->hungry())
		return UseItem(level, SYMBOL_RING);
	    break;
	case 8:
	    if(UseThing() &&
	       (HeavyLoad() || (body->wounded() && rnd.chance(1, 10))))
	    {
		OBJECT* object = pack.FindWorst(this);

		if(object != NULL)
		{
		    OBJECT_LIST list;
		    SPECIFY_OBJECT specify(object);
		    pack.remove(&specify, &list, this, 1L);

		    if(list.length() > 0)
		    {
			MinWorth = max(object->worth(this) + 1, MinWorth);
			BUNCH* bunch = level->create(position());
			bunch->add(&list);
			UseMovePoint(list.length() * 5);

			// Tell the world.
			if(list.length() == 1)
			{
			    list.reset();
			    level->WatchMonster(this, "%s drops %s.",
						list.current());
			}
			else
			    level->WatchMonster(this,
						"%s drops several items.");

			return MayMove() ? CONTINUE : DONE;
		    }
		}
	    }
	    break;
	}
    }

    // * Find nearby player...
    
    if(MayFight())
    {
	SPECIFY_WANT specify(this);
	
	for(region reg(position() - coor(1, 1), position() + coor(2, 2));
	    reg.check();
	    reg.NextX())
	{
	    BUNCH* bunch = level->bunch(reg);
	    
	    if(bunch != NULL && bunch->contain(&specify))
	    {
		if(rnd.chance(2, 3))
		    target = reg;
	    }
	}
    }
    
    // * Find nearest enemy...
    
    if(CanMove())
    {
	if(IsConfused())
	    target = position() + coor(rnd(-50, 50), rnd(-50, 50));
	else if(target == NoWhere || target == position())
	{
	    switch(tactic())
	    {
	    case MOVE_FOLLOW:
		if(rnd.chance(4, 5))
		{
		    PLAYER* player = level->FindPlayer(this);

		    if(player != NULL)
			target = player->position();
		}
		else
		    target = position() + coor(rnd(-50, 50), rnd(-50, 50));

		break;
	    case MOVE_NOWHERE:
		UseMovePoint(50);
		break;
	    case MOVE_INDEPENDENT:
		target = position() + coor(rnd(-50, 50), rnd(-50, 50));
		break;
	    case MOVE_ENEMY:
		if(rnd.chance(4, 5))
		{
		    MONSTER* enemy= level->FindEnemy(this);

		    if(enemy != NULL)
			target = enemy->position();
		}
		else
		    target = position() + coor(rnd(-50, 50), rnd(-50, 50));

		break;
	    case MOVE_END:
	    case MOVE_ERROR:
	    default:
		kill("killed by panic", level);
		bug.error("BEAST::play", format("tactic() == %d", tactic()));
	    }
	}
    }

    // * Special actions

    SpecialAction(level);

    if(!MayMove())
	return DONE;

    // * Try to move
    
    if(target == NoWhere || target == position())
    {
	UseMovePoint(50);
	return DONE;
    }

    straight line(position(), target);
    line.next();
    
    if(IsRock(level->rock(line)))
    {
	if(MayFight() && !weapon()->IsBodyPart() && rnd.chance(250))
	    level->dig(this, line);
	else if(rnd.chance(9))
	    target = NoWhere;
	else
	    target += coor(rnd(-10, 10), rnd(-10, 10));
    }
    else
    {
	BUNCH* from = level->create(position());
	BUNCH* to = level->bunch(line);
	
	if(to == NULL || to->AcceptEnter(this))
	{
	    if(from->AcceptLeave(this) && CanMove())
		move(level, line);
	    else
		UseMovePoint(50);
	}
	else if(MayFight() && CanFight())
	{
	    OBJECT* victim = to->HitOne();

	    if(IsEnemy(victim) || rnd.chance(100))
		to->defend(this, level);
	}
	if(rnd.chance(15))
	    target = NoWhere;
    }
    
    // * Done?
    
    if(MayMove())
	return CONTINUE;

    return DONE;
}

// @@@ void BEAST::save(BINARY* binary)

void BEAST::save(BINARY* binary)
{
    MONSTER::save(binary);
    binary->PutLong(MinWorth);
    enemies.save(binary);
}

// @@@ BEAST::BEAST(BINARY* binary)

BEAST::BEAST(BINARY* binary) : (binary)
{
    MinWorth = binary->GetLong();
    enemies.load(binary);
    target = NoWhere;

    if(enemies.length() > 10)
	enemies.forget(10);
}

// @@@ BEAST::BEAST(PARSE* parse, coor pos)

BEAST::BEAST(PARSE* parse, coor pos) : (parse, pos)
{
    target = NoWhere;
}

// @@@ BEAST::BEAST()

BEAST::BEAST()
{
    target = NoWhere;
}

// @@@ BEAST::~BEAST()

BEAST::~BEAST() { }

// @@ ACID_BLOB

// @@@ void ACID_BLOB::create()

void ACID_BLOB::create()
{
    body = new BODY_ACID_BLOB;
    MyLetter = 'a';
}

// @@ BAT

// @@@ void BAT::create()

void BAT::create()
{
    body = new BODY_BAT;
    SetExperience(2);
    MyLetter = 'b';
}

// @@ RAT

// @@@ void RAT::create(int exp)

void RAT::create(int exp)
{
    body = new BODY_RAT;
    SetExperience(exp);
    MyLetter = 'r';
}

// @@@ bool RAT::IsWanted(OBJECT* object)

bool RAT::IsWanted(OBJECT* object)
{ return object->IsFood(); }

// @@@ int RAT::knowledge(OBJECT* object)

int RAT::knowledge(OBJECT* object)
{
    if(object->IsFood())
	return 50;
    else
	return 0;
}

// @@@ bool RAT::know(OBJECT* object)

bool RAT::know(OBJECT* object)
{
    return object->IsFood();
}

// @@ GOBLIN

// @@@ int GOBLIN::knowledge(OBJECT* object)

int GOBLIN::knowledge(OBJECT* object)
{
    if(object->IsMonster() || object->IsWeapon())
	return 1;
    else if(object->IsFood())
	return 5;
    else
	return 0;
}

// @@@ void GOBLIN::create()

void GOBLIN::create()
{
    body = new BODY_GOBLIN;
    SetExperience(3);
    MyLetter = 'g';
}

// @@ HOBGOBLIN

// @@@ int HOBGOBLIN::knowledge(OBJECT* object)

int HOBGOBLIN::knowledge(OBJECT* object)
{
    if(object->IsMonster() || object->IsWeapon() || object->IsPotion())
	return 15;
    else if(object->IsFood())
	return 10;
    else
	return 0;
}

// @@@ void HOBGOBLIN::create()

void HOBGOBLIN::create()
{
    body = new BODY_HOBGOBLIN;
    SetExperience(10);
    MyLetter = 'H';
}

// @@ ORC

// @@@ void ORC::create()

void ORC::create()
{
    body = new BODY_ORC;
    SetExperience(6);
    MyLetter = 'O';
}

// @@@ bool ORC::IsEnemy(OBJECT* object)

bool ORC::IsEnemy(OBJECT* object)
{
    return (enemies.member(object) ||
	    object->LookType() == MONSTER_ELF ||
	    object->type() == MONSTER_PLAYER);
}

// @@@ int ORC::knowledge(OBJECT* object)

int ORC::knowledge(OBJECT* object)
{
    if(object->IsMonster() || object->IsWeapon() ||
       object->IsArmor() || object->IsFood())
    {
	return 5;
    }
    else
	return 0;
}

// @@ DRAGON

// @@@ void DRAGON::create()

void DRAGON::create()
{
    body = new BODY_DRAGON;
    SetExperience(22);
    MyLetter = 'D';
}

// @@@ void DRAGON::SpecialAction(LEVEL* level)

void DRAGON::SpecialAction(LEVEL* level)
{
    if(MayFight() && rnd.chance(1, body->wounded() ? 5 : 25))
	body->breathe(level, this, enemy(level)->position());
}

// @@@ int DRAGON::knowledge(OBJECT* object)

int DRAGON::knowledge(OBJECT* object)
{
    if(object->IsWand() || object->IsRing())
	return 35;
    else if(object->IsArmor() || object->IsWeapon())
	return 20;
    else
	return 5;
}

// @@@ bool DRAGON::know(OBJECT* object)

bool DRAGON::know(OBJECT* object)
{
    return object->IsWand() || object->IsRing();
}

// @@ LIZARD

// @@@ void LIZARD::create()

void LIZARD::create()
{
    body = new BODY_LIZARD;
    SetExperience(5);
    MyLetter = 'l';
}

// @@ FUNGUS

// @@@ void FUNGUS::create()

void FUNGUS::create()
{
    body = new BODY_FUNGUS;
    SetExperience(10);
    MyLetter = 'f';
}

// @@@ bool FUNGUS::IsEnemy(OBJECT* object)

bool FUNGUS::IsEnemy(OBJECT* object)
{ return object->LookType() != MONSTER_FUNGUS; }

// @@ TROLL

// @@@ void TROLL::create()

void TROLL::create()
{
    body = new BODY_TROLL;
    SetExperience(9);
    MyLetter = 'T';
}

// @@@ int TROLL::knowledge(OBJECT* object)

int TROLL::knowledge(OBJECT* object)
{
    if(object->IsMonster() || object->IsWeapon() || object->IsArmor())
	return 20;
    else if(object->IsFood())
	return 10;
    else
	return 0;
}

// @@ COCKATRICE

// @@@ void COCKATRICE::create()

void COCKATRICE::create()
{
    body = new BODY_COCKATRICE;
    SetExperience(10);
    MyLetter = 'c';
}

// @@@ void COCKATRICE::SpecialAction(LEVEL* level)

void COCKATRICE::SpecialAction(LEVEL* level)
{
    if(MayFight() && rnd.chance(1, body->wounded() ? 10 : 100))
	body->hiss(level, this);
}

// @@ WRAITH

// @@@ void WRAITH::create()

void WRAITH::create()
{
    body = new BODY_WRAITH;
    SetExperience(20);
    MyLetter = 'w';
}

// @@@ bool WRAITH::IsWanted(OBJECT* object)

bool WRAITH::IsWanted(OBJECT* object)
{ return object->IsRing(); }

// @@@ int WRAITH::knowledge(OBJECT* object)

int WRAITH::knowledge(OBJECT* object)
{
    if(object->IsRing())
	return 100;
    else if(object->IsMagic())
	return 30;
    else
	return 0;
}

// @@@ bool WRAITH::know(OBJECT* object)

bool WRAITH::know(OBJECT* object)
{
    return object->IsRing();
}

// @@ ELF

// @@@ void ELF::create()

void ELF::create()
{
    body = new BODY_ELF;
    SetExperience(15);
    MyLetter = 'E';
}

// @@@ bool ELF::IsEnemy(OBJECT* object)

bool ELF::IsEnemy(OBJECT* object)
{ return enemies.member(object); }

// @@@ int ELF::knowledge(OBJECT* object)

int ELF::knowledge(OBJECT* object)
{
    if(object->IsRing() || object->IsWeapon() || object->IsArmor() ||
       object->IsMonster())
    {
	return 50;
    }
    else
	return 0;
}

// @@@ bool ELF::know(OBJECT* object)

bool ELF::know(OBJECT* object)
{
    return object->IsWeapon() || object->IsArmor() || object->IsRing();
}

// @@ WIZARD

// @@@ void WIZARD::create()

void WIZARD::create()
{
    body = new BODY_HUMAN;
    SetExperience(25);
    AddAbility(new COUNTER_TELEPORT_CONTROL);
    AddAbility(new COUNTER_RANDOM_TELEPORT);
    AddAbility(new COUNTER_PHASING);
    AddAbility(new COUNTER_SEE_INVISIBLE);
    AddAbility(new COUNTER_FIRE_PROTECTION);
    AddAbility(new COUNTER_COLD_PROTECTION);
    AddAbility(new COUNTER_ACID_PROTECTION);
    AddAbility(new COUNTER_STONE_PROTECTION);
    AddAbility(new COUNTER_POISON_PROTECTION);
    AddAbility(new COUNTER_HISS);
    AddAbility(new COUNTER_BREATHE_FIRE);
    AddAbility(new COUNTER_HEALTH(0L, 25));
    AddAbility(new COUNTER_FAST(0L, 500));
    body->SetStrength(25);
    MyLetter = 'W';
    pack.add(new AMULET_YENDOR);
}

// @@@ bool WIZARD::IsEnemy(OBJECT* object)

bool WIZARD::IsEnemy(OBJECT* object)
{ return enemies.member(object) || object->type() == MONSTER_DRAGON; }

// @@@ const char* WIZARD::name()

const char* WIZARD::name()
{
    if(dead())
	return "the masters dead body";
    else
	return "The Wizard of Yendor";
}

// @@@ void WIZARD::SpecialAction(LEVEL* level)

void WIZARD::SpecialAction(LEVEL* level)
{
    if(body->wounded() && MayFight())
    {
	switch(rnd(1, 15))
	{
	case 1:
	    body->breathe(level, this, enemy(level)->position());
	    break;
	case 2:
	    body->hiss(level, this);
	    break;
	}
    }
    else
    {
	PLAYER* player = level->FindPlayer(this);
	
	if(player != NULL && !IsEnemy(player))
	{
	    coor diff = abs(position() - player->position());
	    
	    if(diff.x < 2 && diff.y < 2)
	    {
		if(!talker.member(player))
		{
		    level->talk(this, " Killed a dragon?");
		    talker.add(player);
		}
		else if(amulet.member(player))
		    /* Do nothing */;
		else if(player->DragonKiller())
		{
		    level->talk(this, " You did it!");
		    level->talk(this, " Have this, hero.");
		    player->message("You feel magic in your knapsack.");
		    player->pack.add(new AMULET_YENDOR);
		    amulet.add(player);
		}
	    }
	}
    }
}

// @@@ void WIZARD::save(BINARY* binary)

void WIZARD::save(BINARY* binary)
{
    BEAST::save(binary);
    talker.save(binary);
    amulet.save(binary);
}

// @@@ WIZARD::WIZARD(BINARY* binary)

WIZARD::WIZARD(BINARY* binary) : (binary)
{
    talker.load(binary);
    amulet.load(binary);

    if(talker.length() > 50)
	talker.forget(25);
    if(amulet.length() > 1000)
	amulet.forget(100);
}

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
