// @ myth - server - object.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ include

#include "server/object.h"
#include "common/error.h"
#include "common/binary.h"
#include "common/parse.h"
#include "server/rnd.h"
#include "server/weapon.h"
#include "server/armor.h"
#include "server/food.h"
#include "server/player.h"
#include "server/potion.h"
#include "server/scroll.h"
#include "server/wand.h"
#include "server/ring.h"
#include "server/stair.h"
#include "server/beast.h"
#include "common/format.h"

// @@ object_type ObjectToType(const char* string)

// @@@ static const struct ObjectTable[]

static const struct
{
    const char* string;
    object_type value;
    int chance;
} ObjectTable[] = {
    "STAIR", OBJECT_STAIR, 0,
    "POTION_WATER", POTION_OF_WATER, 25,
    "POTION_STRENGTH", POTION_OF_STRENGTH, 25,
    "POTION_HEALING", POTION_OF_HEALING, 25,
    "POTION_EXPERIENCE", POTION_OF_EXPERIENCE, 25,
    "POTION_SIGHT", POTION_OF_SIGHT, 25,
    "POTION_ENERGY", POTION_OF_ENERGY, 25,
    "SCROLL_BLANK_PAPER", SCROLL_OF_BLANK_PAPER, 50,
    "SCROLL_KNOWLEDGE", SCROLL_OF_KNOWLEDGE, 100,
    "SCROLL_PRAYERS", SCROLL_OF_PRAYERS, 50,
    "SCROLL_MAGIC", SCROLL_OF_MAGIC, 50,
    "SCROLL_TELEPORT", SCROLL_OF_TELEPORT, 50,
    "SCROLL_MAPPING", SCROLL_OF_MAPPING, 50,
    "TORCH", WAND_TORCH, 250,
    "WAND_SLEEP", WAND_OF_SLEEP, 10,
    "WAND_LIFE", WAND_OF_LIFE, 10,
    "WAND_TELEPORT", WAND_OF_TELEPORT, 10,
    "WAND_VISIBILITY", WAND_OF_VISIBILITY, 10,
    "WAND_MISSILE", WAND_OF_MISSILE, 10,
    "WAND_LIGHT", WAND_OF_LIGHT, 10,
    "RING_ADORNMENT", RING_OF_ADORNMENT, 5,
    "RING_HEALTH", RING_OF_HEALTH, 5,
    "RING_TELEPORTATION", RING_OF_TELEPORTATION, 5,
    "RING_VISION", RING_OF_VISION, 5,
    "RING_STRENGTH", RING_OF_STRENGTH, 5,
    "RING_FIRE_RESISTANCE", RING_OF_FIRE_RESISTANCE, 5,
    "RING_COLD_RESISTANCE", RING_OF_COLD_RESISTANCE, 5,
    "AMULET_YENDOR", AMULET_OF_YENDOR, 0,
    "STONE", WEAPON_STONE, 500,
    "BOW", WEAPON_BOW, 10,
    "ARROW", WEAPON_ARROW, 50,
    "DAGGER", WEAPON_DAGGER, 15,
    "MACE", WEAPON_MACE, 50,
    "SWORD", WEAPON_LONG_SWORD, 25,
    "TWO_HANDED_SWORD", WEAPON_TWO_HANDED_SWORD, 10,
    "LEATHER_ARMOR", ARMOR_LEATHER, 50,
    "RING_MAIL", ARMOR_RING, 25,
    "PLATE_MAIL", ARMOR_PLATE, 10,
    "RATION", FOOD_RATION, 250,
    "FRUIT", FOOD_FRUIT, 500,
    "PLAYER", MONSTER_PLAYER, 0,
    "WIZARD", MONSTER_WIZARD, 0,
    "ACID_BLOB", MONSTER_ACID_BLOB, 0,
    "BAT", MONSTER_BAT, 0,
    "RAT", MONSTER_RAT, 0,
    "GOBLIN", MONSTER_GOBLIN, 0,
    "HOBGOBLIN", MONSTER_HOBGOBLIN, 0,
    "ORC", MONSTER_ORC, 0,
    "DRAGON", MONSTER_DRAGON, 0,
    "LIZARD", MONSTER_LIZARD, 0,
    "COCKATRICE", MONSTER_COCKATRICE, 0,
    "WRAITH", MONSTER_WRAITH, 0,
    "ELF", MONSTER_ELF, 0,
    "TROLL", MONSTER_TROLL, 0,
    "FUNGUS", MONSTER_FUNGUS, 0,
    "END", OBJECT_END, 0,
    NULL, OBJECT_ERROR, 0,
};

// @@@ object_type ObjectToType(const char* string)

object_type ObjectToType(const char* string)
{
    for(int i = 0; ObjectTable[i].string != NULL; i++)
    {
	if(strcmp(string, ObjectTable[i].string) == EQUAL)
	    return ObjectTable[i].value;
    }
    return OBJECT_ERROR;
}

// @@@ OBJECT* ::CreateThing(int value)

OBJECT* CreateThing(int value)
{
    static int sum = 0;		// Sum of chance for alll objects
    object_type type;		// Type of object to create
     
    // Initialise this function...

    if(sum == 0)
    {
	for(type = 0; type < OBJECT_END; type++)
	    sum += ObjectTable[type].chance;
    }

    // Find out which object to create
    
    int where = rnd(1, sum);

    for(type = 0; where > ObjectTable[type].chance; type++)
	where -= ObjectTable[type].chance;

    OBJECT* object = CreateObject(type);

    // Things can be blessed and cursed
    
    if(rnd.chance(value, 250))
	object->bless();
    else if(rnd.chance(value, 100))
	object->curse();
    
    // And thing have a value..
    
    if(rnd.chance(2))
	object->ChangeValue(rnd(1, value));
    else if(rnd.chance(5))
	object->ChangeValue(rnd(1, 25));
    else 
	object->ChangeValue(value);

    return object;
}    

// @@ OBJECT

// @@@ bool OBJECT::match(symbol_type type, char c)

bool OBJECT::match(symbol_type type, char c)

{
    if(type == SYMBOL_BUNCH && IsKind(symbol()))
	return TRUE;
    else if(type != symbol())
	return FALSE;
    else if(c == SYMBOL_BUNCH)
	return TRUE;
    else if(c != letter())
	return FALSE;
    else
	return TRUE;

    // I am paid per line...
}

// @@@ comp OBJECT::compare(OBJECT* object)

comp OBJECT::compare(OBJECT* object)
{ return ::compare(type(), object->type()); }

// @@@ STRING OBJECT::print(print_flag flag, long count, PLAYER*)

STRING OBJECT::print(print_flag flag, long count, PLAYER*)
{
    if((flag & PRINT_TYPE))
    {
	if(count == 1L)
	    return format("%s", name());

	return format("%s", PluralName());
    }
    if(count == 1L)
	return format("%s", ClassName());

    return format("%s", (PluralClassName()));
}

// @@@ const char* OBJECT::name()

const char* OBJECT::name()
{
    bug.error("OBJECT::name", "virtual");
    return "bug";
}

// @@@ const char* OBJECT::PluralName()

const char* OBJECT::PluralName()
{
    STRING string = format("%ss", name());
    const char* text = string.string();
    return text; 
}

// @@@ const char* OBJECT::ClassName()

const char* OBJECT::ClassName()
{ return name(); }

// @@@ const char* OBJECT::PluralClassName()

const char* OBJECT::PluralClassName()
{
    STRING string = format("%ss", ClassName());
    const char* text = string.string();
    return text;
}

// @@@ void OBJECT::apply(LEVEL*, MONSTER* monster)

void OBJECT::apply(LEVEL*, MONSTER* monster)
{
    monster->message(format("You can't figure out how to apply %s.",
			  ::print(this, PRINT_OBJECT)));
    monster->pack.add(this);
}

// @@@ QUESTION* OBJECT::DoApply(LEVEL* level, PLAYER* player, CHANNEL*)

QUESTION* OBJECT::DoApply(LEVEL* level, PLAYER* player, CHANNEL*)
{
    apply(level, player);
    return NULL;
}

// @@@ bool OBJECT::MoveUp(LEVEL*, MONSTER*)

bool OBJECT::MoveUp(LEVEL*, MONSTER*)
{ return FALSE; }

// @@@ bool OBJECT::MoveDown(LEVEL*, MONSTER*)

bool OBJECT::MoveDown(LEVEL*, MONSTER*)
{ return FALSE; }

// @@@ bool OBJECT::AcceptEnter(MONSTER*)

bool OBJECT::AcceptEnter(MONSTER*)
{ return TRUE; }

// @@@ bool OBJECT::AcceptLeave(MONSTER*)

bool OBJECT::AcceptLeave(MONSTER*)
{ return TRUE; }

// @@@ QUESTION* OBJECT::GetQuestion(PLAYER*, LEVEL*, coor)

QUESTION* OBJECT::GetQuestion(PLAYER*, LEVEL*, coor)
{ return NULL; }

// @@@ cont OBJECT::defend(MONSTER*, LEVEL*)

cont OBJECT::defend(MONSTER*, LEVEL*)
{ return CONTINUE; }

// @@@ void OBJECT::MoveAway(LEVEL* level, coor pos, int)
//
// Can't teleport a monster the same way as an object

void OBJECT::MoveAway(LEVEL* level, coor pos, int)
{
    BUNCH* bunch = level->create(pos);
    bunch->add(this);
}

// @@@ bool OBJECT::MightHit()
//
// Return TRUE, iff this object is is a candidate for a blow by an
// orc's two handed sword, or the like.

bool OBJECT::MightHit()
{
    if(size() > 1000)
	return !rnd.chance((int) size() / 250);
    else
	return rnd.chance((int) size(), 4000);
}

// @@@ bool OBJECT::catch(int, OBJECT*)
//
// Most objects can't catch anything

bool OBJECT::catch(int, OBJECT*)
{ return FALSE; }

// @@@ bool OBJECT::hit(int, OBJECT*)
//
// Check to see if we hit

bool OBJECT::hit(int, OBJECT*, MONSTER*)
{ return rnd.chance((int) size(), 1000); }

// @@@ STRING OBJECT::hurt(int power, OBJECT* object, MONSTER*, LEVEL*)
//
// We *did* hit, now hurt it!

STRING OBJECT::hurt(int power, OBJECT* object, MONSTER*, LEVEL*)
{
    // Damage it
    if(rnd.chance(power, power + strength() * 5))
	damage();
    else if(rnd.chance(power, power + object->strength() *5))
	object->damage();

    return NullString;
}

// @@@ void OBJECT::SetLetter(char)

void OBJECT::SetLetter(char)
{ /* Do Nothing */ }

// @@@ void OBJECT::add(OBJECT*)

void OBJECT::add(OBJECT*)
{ bug.error("OBJECT::add", "virtual"); }

// @@@ OBJECT* OBJECT::split(long)

OBJECT* OBJECT::split(long)
{ return NULL; }

// @@@ void OBJECT::save(BINARY* binary)

void OBJECT::save(BINARY* binary)
{
    // We must save the type first, in order to be able to restore it
    binary->PutShort(type()); // See also OBJECT_LIST::save(BINARY*)
    // We don't know how to save id_type
    SaveIdentity(binary, id);
}

// @@@ OBJECT::OBJECT(BINARY* binary)

OBJECT::OBJECT(BINARY* binary)
{
    id = LoadIdentity(binary);
}

// @@@ OBJECT::OBJECT(PARSE*)

OBJECT::OBJECT(PARSE*)
{ id = ::identity(); }

// @@@ OBJECT::OBJECT()

OBJECT::OBJECT()
{ id = ::identity(); }

// @@@ OBJECT* ::ParseObject(PARSE* parse, coor pos)

OBJECT* ParseObject(PARSE* parse, coor pos)
{
    const char* string = parse->GetWord();
    parse->skip(INPUT_COMMENT);

    switch(ObjectToType(string))
    {
    case OBJECT_STAIR:
	return new STAIR(parse);
    case POTION_OF_WATER:
	return new POTION_WATER(parse);
    case POTION_OF_STRENGTH:
	return new POTION_STRENGTH(parse);
    case POTION_OF_HEALING:
	return new POTION_HEALING(parse);
    case POTION_OF_EXPERIENCE:
	return new POTION_EXPERIENCE(parse);
    case POTION_OF_SIGHT:
	return new POTION_SIGHT(parse);
    case POTION_OF_ENERGY:
	return new POTION_ENERGY(parse);
    case SCROLL_OF_BLANK_PAPER:
	return new SCROLL_BLANK_PAPER(parse);
    case SCROLL_OF_KNOWLEDGE:
	return new SCROLL_KNOWLEDGE(parse);
    case SCROLL_OF_PRAYERS:
	return new SCROLL_PRAYERS(parse);
    case SCROLL_OF_MAGIC:
	return new SCROLL_MAGIC(parse);
    case SCROLL_OF_TELEPORT:
	return new SCROLL_TELEPORT(parse);
    case SCROLL_OF_MAPPING:
	return new SCROLL_MAPPING(parse);
    case WAND_TORCH:
	return new TORCH(parse);
    case WAND_OF_SLEEP:
	return new WAND_SLEEP(parse);
    case WAND_OF_LIFE:
	return new WAND_LIFE(parse);
    case WAND_OF_TELEPORT:
	return new WAND_TELEPORT(parse);
    case WAND_OF_VISIBILITY:
	return new WAND_VISIBILITY(parse);
    case WAND_OF_MISSILE:
	return new WAND_MISSILE(parse);
    case WAND_OF_LIGHT:
	return new WAND_LIGHT(parse);
    case RING_OF_ADORNMENT:
	return new RING_ADORNMENT(parse);
    case RING_OF_HEALTH:
	return new RING_HEALTH(parse);
    case RING_OF_TELEPORTATION:
	return new RING_TELEPORTATION(parse);
    case RING_OF_VISION:
	return new RING_VISION(parse);
    case RING_OF_STRENGTH:
	return new RING_STRENGTH(parse);
    case RING_OF_FIRE_RESISTANCE:
	return new RING_FIRE_RESISTANCE(parse);
    case RING_OF_COLD_RESISTANCE:
	return new RING_COLD_RESISTANCE(parse);
    case AMULET_OF_YENDOR:
	return new AMULET_YENDOR(parse);
    case WEAPON_STONE:
	return new STONE(parse);
    case WEAPON_BOW:
	return new BOW(parse);
    case WEAPON_ARROW:
	return new ARROW(parse);
    case WEAPON_DAGGER:
	return new DAGGER(parse);
    case WEAPON_MACE:
	return new MACE(parse);
    case WEAPON_LONG_SWORD:
	return new SWORD(parse);
    case WEAPON_TWO_HANDED_SWORD:
	return new TWO_HANDED_SWORD(parse);
    case ARMOR_LEATHER:
	return new LEATHER_ARMOR(parse);
    case ARMOR_RING:
	return new RING_MAIL(parse);
    case ARMOR_PLATE:
	return new PLATE_MAIL(parse);
    case FOOD_RATION:
	return new RATION(parse);
    case FOOD_FRUIT:
	return new FRUIT(parse);
    case MONSTER_PLAYER:
	return new PLAYER(parse, pos);
    case MONSTER_WIZARD:
	return new WIZARD(parse, pos);
    case MONSTER_ACID_BLOB:
	return new ACID_BLOB(parse, pos);
    case MONSTER_BAT:
	return new BAT(parse, pos);
    case MONSTER_RAT:
	return new RAT(parse, pos);
    case MONSTER_GOBLIN:
	return new GOBLIN(parse, pos);
    case MONSTER_HOBGOBLIN:
	return new HOBGOBLIN(parse, pos);
    case MONSTER_ORC:
	return new ORC(parse, pos);
    case MONSTER_DRAGON:
	return new DRAGON(parse, pos);
    case MONSTER_LIZARD:
	return new LIZARD(parse, pos);
    case MONSTER_COCKATRICE:
	return new COCKATRICE(parse, pos);
    case MONSTER_WRAITH:
	return new WRAITH(parse, pos);
    case MONSTER_ELF:
	return new ELF(parse, pos);
    case MONSTER_TROLL:
	return new TROLL(parse, pos);
    case MONSTER_FUNGUS:
	return new FUNGUS(parse, pos);
    case OBJECT_END:
	return NULL;
    case OBJECT_ERROR:
    default:
	parse->wrong("OBJECT_LIST::OBJECT_LIST",
		     format("Unknown object %s", string));
	return NULL;
    }
}

// @@@ OBJECT* ::LoadObject(BINARY* binary)

OBJECT* LoadObject(BINARY* binary)
{
    switch(binary->GetShort())
    {
    case OBJECT_STAIR:
	return new STAIR(binary);
    case POTION_OF_WATER:
	return new POTION_WATER(binary);
    case POTION_OF_STRENGTH:
	return new POTION_STRENGTH(binary);
    case POTION_OF_HEALING:
	return new POTION_HEALING(binary);
    case POTION_OF_EXPERIENCE:
	return new POTION_EXPERIENCE(binary);
    case POTION_OF_SIGHT:
	return new POTION_SIGHT(binary);
    case POTION_OF_ENERGY:
	return new POTION_ENERGY(binary);
    case SCROLL_OF_BLANK_PAPER:
	return new SCROLL_BLANK_PAPER(binary);
    case SCROLL_OF_KNOWLEDGE:
	return new SCROLL_KNOWLEDGE(binary);
    case SCROLL_OF_PRAYERS:
	return new SCROLL_PRAYERS(binary);
    case SCROLL_OF_MAGIC:
	return new SCROLL_MAGIC(binary);
    case SCROLL_OF_TELEPORT:
	return new SCROLL_TELEPORT(binary);
    case SCROLL_OF_MAPPING:
	return new SCROLL_MAPPING(binary);
    case WAND_TORCH:
	return new TORCH(binary);
    case WAND_OF_SLEEP:
	return new WAND_SLEEP(binary);
    case WAND_OF_LIFE:
	return new WAND_LIFE(binary);
    case WAND_OF_TELEPORT:
	return new WAND_TELEPORT(binary);
    case WAND_OF_VISIBILITY:
	return new WAND_VISIBILITY(binary);
    case WAND_OF_MISSILE:
	return new WAND_MISSILE(binary);
    case WAND_OF_LIGHT:
	return new WAND_LIGHT(binary);
    case RING_OF_ADORNMENT:
	return new RING_ADORNMENT(binary);
    case RING_OF_HEALTH:
	return new RING_HEALTH(binary);
    case RING_OF_TELEPORTATION:
	return new RING_TELEPORTATION(binary);
    case RING_OF_VISION:
	return new RING_VISION(binary);
    case RING_OF_STRENGTH:
	return new RING_STRENGTH(binary);
    case RING_OF_FIRE_RESISTANCE:
	return new RING_FIRE_RESISTANCE(binary);
    case RING_OF_COLD_RESISTANCE:
	return new RING_COLD_RESISTANCE(binary);
    case AMULET_OF_YENDOR:
	return new AMULET_YENDOR(binary);
    case WEAPON_STONE:
	return new STONE(binary);
    case WEAPON_BOW:
	return new BOW(binary);
    case WEAPON_ARROW:
	return new ARROW(binary);
    case WEAPON_DAGGER:
	return new DAGGER(binary);
    case WEAPON_MACE:
	return new MACE(binary);
    case WEAPON_LONG_SWORD:
	return new SWORD(binary);
    case WEAPON_TWO_HANDED_SWORD:
	return new TWO_HANDED_SWORD(binary);
    case ARMOR_LEATHER:
	return new LEATHER_ARMOR(binary);
    case ARMOR_RING:
	return new RING_MAIL(binary);
    case ARMOR_PLATE:
	return new PLATE_MAIL(binary);
    case FOOD_RATION:
	return new RATION(binary);
    case FOOD_FRUIT:
	return new FRUIT(binary);
    case MONSTER_PLAYER:
	return new PLAYER(binary);
    case MONSTER_WIZARD:
	return new WIZARD(binary);
    case MONSTER_ACID_BLOB:
	return new ACID_BLOB(binary);
    case MONSTER_BAT:
	return new BAT(binary);
    case MONSTER_RAT:
	return new RAT(binary);
    case MONSTER_GOBLIN:
	return new GOBLIN(binary);
    case MONSTER_HOBGOBLIN:
	return new HOBGOBLIN(binary);
    case MONSTER_ORC:
	return new ORC(binary);
    case MONSTER_DRAGON:
	return new DRAGON(binary);
    case MONSTER_LIZARD:
	return new LIZARD(binary);
    case MONSTER_COCKATRICE:
	return new COCKATRICE(binary);
    case MONSTER_WRAITH:
	return new WRAITH(binary);
    case MONSTER_ELF:
	return new ELF(binary);
    case MONSTER_TROLL:
	return new TROLL(binary);
    case MONSTER_FUNGUS:
	return new FUNGUS(binary);
    case OBJECT_END:
	return NULL;
    case OBJECT_ERROR:
    default:
	bug.warning("::LoadObject", "Unknown object");
	return NULL;
    }
}

// @@@ OBJECT* ::CreateObject(object_type type)

OBJECT* CreateObject(object_type type)
{
    switch(type)
    {
    case POTION_OF_WATER:
	return new POTION_WATER;
    case POTION_OF_STRENGTH:
	return new POTION_STRENGTH;
    case POTION_OF_HEALING:
	return new POTION_HEALING;
    case POTION_OF_EXPERIENCE:
	return new POTION_EXPERIENCE;
    case POTION_OF_SIGHT:
	return new POTION_SIGHT;
    case POTION_OF_ENERGY:
	return new POTION_ENERGY;
    case SCROLL_OF_BLANK_PAPER:
	return new SCROLL_BLANK_PAPER;
    case SCROLL_OF_KNOWLEDGE:
	return new SCROLL_KNOWLEDGE;
    case SCROLL_OF_PRAYERS:
	return new SCROLL_PRAYERS;
    case SCROLL_OF_MAGIC:
	return new SCROLL_MAGIC;
    case SCROLL_OF_TELEPORT:
	return new SCROLL_TELEPORT;
    case SCROLL_OF_MAPPING:
	return new SCROLL_MAPPING;
    case WAND_TORCH:
	return new TORCH;
    case WAND_OF_SLEEP:
	return new WAND_SLEEP;
    case WAND_OF_LIFE:
	return new WAND_LIFE;
    case WAND_OF_TELEPORT:
	return new WAND_TELEPORT;
    case WAND_OF_VISIBILITY:
	return new WAND_VISIBILITY;
    case WAND_OF_MISSILE:
	return new WAND_MISSILE;
    case WAND_OF_LIGHT:
	return new WAND_LIGHT;
    case RING_OF_ADORNMENT:
	return new RING_ADORNMENT;
    case RING_OF_HEALTH:
	return new RING_HEALTH;
    case RING_OF_TELEPORTATION:
	return new RING_TELEPORTATION;
    case RING_OF_VISION:
	return new RING_VISION;
    case RING_OF_STRENGTH:
	return new RING_STRENGTH;
    case RING_OF_FIRE_RESISTANCE:
	return new RING_FIRE_RESISTANCE;
    case RING_OF_COLD_RESISTANCE:
	return new RING_COLD_RESISTANCE;
    case AMULET_OF_YENDOR:
	return new AMULET_YENDOR;
    case WEAPON_STONE:
	return new STONE;
    case WEAPON_BOW:
	return new BOW;
    case WEAPON_ARROW:
	return new ARROW;
    case WEAPON_DAGGER:
	return new DAGGER;
    case WEAPON_MACE:
	return new MACE;
    case WEAPON_LONG_SWORD:
	return new SWORD;
    case WEAPON_TWO_HANDED_SWORD:
	return new TWO_HANDED_SWORD;
    case ARMOR_LEATHER:
	return new LEATHER_ARMOR;
    case ARMOR_RING:
	return new RING_MAIL;
    case ARMOR_PLATE:
	return new PLATE_MAIL;
    case FOOD_RATION:
	return new RATION;
    case FOOD_FRUIT:
	return new FRUIT;
    case MONSTER_WIZARD:
	return new WIZARD;
    case MONSTER_ACID_BLOB:
	return new ACID_BLOB;
    case MONSTER_BAT:
	return new BAT;
    case MONSTER_RAT:
	return new RAT;
    case MONSTER_GOBLIN:
	return new GOBLIN;
    case MONSTER_HOBGOBLIN:
	return new HOBGOBLIN;
    case MONSTER_ORC:
	return new ORC;
    case MONSTER_DRAGON:
	return new DRAGON;
    case MONSTER_LIZARD:
	return new LIZARD;
    case MONSTER_COCKATRICE:
	return new COCKATRICE;
    case MONSTER_WRAITH:
	return new WRAITH;
    case MONSTER_ELF:
	return new ELF;
    case MONSTER_TROLL:
	return new TROLL;
    case MONSTER_FUNGUS:
	return new FUNGUS;
    case OBJECT_END:
	return NULL;
    case MONSTER_PLAYER:
    case OBJECT_STAIR:
    case OBJECT_ERROR:
    default:
	bug.warning("::CreateObject", format("Illegal object %d", type));
	return NULL;
    }
}

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
